# Copyright (C) 2018 Cancer Care Associates

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.

import numpy as np

from scipy.optimize import basinhopping
from scipy.interpolate import splprep, splev

import pydicom

from pymedphys_utilities.libutils import get_imports
from pymedphys_analysis.geometry import (
    cubify_cube_definition, cube_vertices, cube_vectors)

IMPORTS = get_imports(globals())


# pylint: disable=C0103

def pull_structure_by_number(number, dcm_struct):
    contours_by_slice_raw = [
        item.ContourData
        for item in dcm_struct.ROIContourSequence[number].ContourSequence
    ]
    x = [
        np.array(item[0::3])
        for item in contours_by_slice_raw]
    y = [
        np.array(item[1::3])
        for item in contours_by_slice_raw]
    z = [
        np.array(item[2::3])
        for item in contours_by_slice_raw]

    return x, y, z


def pull_structure(string, dcm_struct):
    structure_names = np.array(
        [item.ROIName for item in dcm_struct.StructureSetROISequence])
    reference = structure_names == string
    if np.all(reference == False):  # pylint: disable=C0121
        raise Exception("Structure not found (case sensitive)")

    index = int(np.where(reference)[0])
    x, y, z = pull_structure_by_number(
        index, dcm_struct)

    return x, y, z


def list_structures(dcm_struct):
    return [item.ROIName for item in dcm_struct.StructureSetROISequence]


def resample_contour(contour, n=50):
    tck, u = splprep(
        [contour[0], contour[1], contour[2]], s=0, k=1)
    new_points = splev(np.arange(0, 1, 1/n), tck)

    return new_points


def resample_contour_set(contours, n=50):

    resampled_contours = [
        resample_contour([x, y, z], n)
        for x, y, z in zip(*contours)
    ]

    return resampled_contours


def contour_to_points(contours):
    resampled_contours = resample_contour_set([
        contours[1], contours[0], contours[2]])
    contour_points = np.concatenate(resampled_contours, axis=1)

    return contour_points


def get_structure_aligned_cube(x0: np.ndarray, structure_name: str,
                               dcm_struct: pydicom.dataset.FileDataset,
                               quiet=False, niter=10):
    """Align a cube to a dicom structure set.

    Designed to allow arbitrary references frames within a dicom file to be
    extracted via contouring a cube.

    Parameters
    ----------
    x0
        A 3x3 array with each row defining a 3-D point in space.
        These three points are used as initial conditions to search for
        a cube that fits the contours. Choosing initial values close to the
        structure set, and in the desired orientation will allow consistent
        results. See examples within
        `pymedphys.geometry.cubify_cube_definition`_ on what the effects
        of each of the three points are on the resulting cube.
    structure_name
        The DICOM label of the cube structure
    dcm_struct
        The pydicom reference to the DICOM structure file.
    quiet : bool
        Tell the function to not print anything. Defaults to False.

    Returns
    -------
    cube_definition_array
        Four 3-D points the define the vertices of the cube.

    vectors
        The vectors between the points that can be used to traverse the cube.

    Examples
    --------
    >>> import numpy as np
    >>> import pydicom
    >>> from pymedphys.geometry import get_structure_aligned_cube
    >>>
    >>> struct_path = 'packages/pymedphys_dicom/tests/data/struct/example_structures.dcm'
    >>> dcm_struct = pydicom.dcmread(struct_path, force=True)
    >>>
    >>> x0 = np.array([
    ...     -270, -35, -20,
    ...     -270, 30, -20,
    ...     -260, -35, 40
    ... ])
    >>>
    >>> structure_name = 'ANT Box'
    >>> cube_definition_array, vectors = get_structure_aligned_cube(
    ...     x0, structure_name, dcm_struct, quiet=True, niter=1)
    >>> np.round(cube_definition_array)
    array([[-276.,  -31.,  -16.],
           [-275.,   29.,  -17.],
           [-266.,  -31.,   43.],
           [-217.,  -32.,  -26.]])
    >>>
    >>> np.round(vectors, 1)
    array([[ 0.7, 59.9, -0.5],
           [ 9.7,  0.4, 59.1],
           [59.1, -0.8, -9.7]])
    """

    to_minimise = create_minimise(structure_name, dcm_struct)

    if quiet:
        def print_fun(x, f, accepted):
            pass
    else:
        def print_fun(x, f, accepted):
            print("at minimum %.4f accepted %d" % (f, int(accepted)))

    result = basinhopping(
        to_minimise, x0, callback=print_fun, niter=niter, stepsize=5)

    cube = result.x

    cube_definition = cubify_cube_definition(
        [tuple(cube[0:3]), tuple(cube[3:6]), tuple(cube[6::])]
    )

    cube_definition_array = [
        np.array(list(item))
        for item in cube_definition
    ]

    vectors = [
        cube_definition_array[1] - cube_definition_array[0],
        cube_definition_array[2] - cube_definition_array[0],
        cube_definition_array[3] - cube_definition_array[0]
    ]

    return cube_definition_array, vectors


def calc_min_distance(cube_definition, contours):
    vertices = cube_vertices(cube_definition)

    vectors = cube_vectors(cube_definition)
    unit_vectors = [
        vector / np.linalg.norm(vector)
        for vector in vectors
    ]

    plane_norms = np.array([
        unit_vectors[1],
        -unit_vectors[0],
        -unit_vectors[1],
        unit_vectors[0],
        unit_vectors[2],
        -unit_vectors[2]
    ])

    plane_points = np.array([
        vertices[0],
        vertices[1],
        vertices[2],
        vertices[0],
        vertices[0],
        vertices[3]
    ])

    plane_origin_dist = -np.sum(plane_points * plane_norms, axis=1)

    distance_to_planes = np.dot(
        plane_norms, contours) + plane_origin_dist[:, None]
    min_dist_squared = np.min(distance_to_planes**2, axis=0)

    return min_dist_squared


def create_minimise(structure_name, dcm_struct):
    contours = pull_structure(structure_name, dcm_struct)
    contour_points = contour_to_points(contours)

    def minimise(cube):
        cube_definition = cubify_cube_definition(
            [tuple(cube[0:3]), tuple(cube[3:6]), tuple(cube[6::])]
        )
        min_dist_squared = calc_min_distance(cube_definition, contour_points)
        return np.sum(min_dist_squared)

    return minimise



# Copyright (C) 2018 Simon Biggs

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.

"""A DICOM toolbox. Available functions include:

>>> from pymedphys.dicom import (
...
...     # General functions
...     anonymise_dataset,
...     anonymise_file,
...     anonymise_directory,
...     is_anonymised_dataset,
...     is_anonymised_file,
...     is_anonymised_directory,
...
...     # RT Dose related functions
...     extract_iec_patient_xyz,
...     extract_iec_fixed_xyz,
...     extract_dicom_patient_xyz,
...
...     # RT Structure related functions
...     get_structure_aligned_cube,
...
...     # Functions using a combination of DICOM Modalities
...     find_dose_within_structure,
...     create_dvh)
"""

# pylint: disable=W0401,W0614

from pymedphys_utilities.libutils import clean_and_verify_levelled_modules

from ._level1.constants import *
from ._level1.structure import *
from ._level1.create import *
from ._level2.header_tweaks import *
from ._level2.anonymise import *
from ._level2.dose import *
from ._level3.dicom_collection import *


clean_and_verify_levelled_modules(globals(), [
    '._level1.structure', '._level1.create', '._level2.header_tweaks',
    '._level2.anonymise', '._level2.dose', '._level3.dicom_collection'
], package='pymedphys_dicom.dicom')



# Copyright (C) 2019 Simon Biggs, Matt Jennings

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.


from copy import deepcopy
from packaging import version

import numpy as np
import pydicom

from .._level1.create import dicom_dataset_from_dict
from .._level2.anonymise import anonymise_dataset
from .._level2.dose import (
    extract_dicom_patient_xyz, convert_xyz_to_dicom_coords)

from pymedphys_utilities.libutils import get_imports

IMPORTS = get_imports(globals())

# pylint: disable=W0201


# Many thanks to the Medical Connections for offering free
# valid UIDs (http://www.medicalconnections.co.uk/FreeUID.html)
# Their service was used to obtain the following root UID for PyMedPhys:
PYMEDPHYS_ROOT_UID = '1.2.826.0.1.3680043.10.188'


class DicomBase:
    def __init__(self, dataset, copy=True):
        if copy:
            dataset = deepcopy(dataset)

        if dataset.is_little_endian is None:
            dataset.is_little_endian = True

        if dataset.is_implicit_VR is None:
            dataset.is_implicit_VR = True

        self.dataset = dataset

    @classmethod
    def from_file(cls, fp):
        """Instantiate a DicomBase instance from a filepath or file-like
        object.
        """
        dataset = pydicom.dcmread(fp, force=True)

        return cls(dataset)

    def to_file(self, fp):
        """Write the DicomBase instance's dataset to a file denoted by
        a filepath or file-like object.
        """
        self.dataset.save_as(fp)

    @classmethod
    def from_dict(cls, dictionary):
        """Instantiate a DicomBase instance from a dictionary of
        DICOM keyword/value pairs.
        """
        dataset = dicom_dataset_from_dict(dictionary)

        return cls(dataset)


    def _pydicom_eq(self, other):
        return self.dataset == other.dataset


    def __repr__(self):
        return self.dataset.__repr__()


    def __eq__(self, other):
        if version.parse(pydicom.__version__) <= version.parse("1.2.1"):
            self_elems = sorted(list(self.dataset.iterall()),
                                key=lambda x: x.tag)
            other_elems = sorted(list(other.dataset.iterall()),
                                 key=lambda x: x.tag)
            return self_elems == other_elems
        else:
            # TODO: Change for pydicom>=1.2.2?
            self_elems = sorted(list(self.dataset.iterall()),
                                key=lambda x: x.tag)
            other_elems = sorted(list(other.dataset.iterall()),
                                 key=lambda x: x.tag)
            return self_elems == other_elems

    def __ne__(self, other):
        return not self == other


    def anonymise(self, inplace=False):
        to_copy = not inplace
        anonymised = anonymise_dataset(self.dataset, copy_dataset=to_copy)

        if not inplace:
            return self.__class__(anonymised)


class DicomDose(DicomBase):

    def __init__(self, dataset, copy=True):
        super().__init__(dataset, copy=copy)

        self.mask = None

    @property
    def values(self) -> np.ndarray:
        return self.dataset.pixel_array * self.dataset.DoseGridScaling

    @property
    def units(self):
        return self.dataset.DoseUnits

    # TODO: Need to refactor the following to still be a view to the dataset
    # but not needlessly call the entire function.
    @property
    def x(self):
        x_value, _, _ = extract_dicom_patient_xyz(self.dataset)
        return x_value

    @property
    def y(self):
        _, y_value, _ = extract_dicom_patient_xyz(self.dataset)
        return y_value

    @property
    def z(self):
        _, _, z_value = extract_dicom_patient_xyz(self.dataset)
        return z_value

    @property
    def coords(self):
        x, y, z = extract_dicom_patient_xyz(self.dataset)
        return convert_xyz_to_dicom_coords((x, y, z))


class DicomImage(DicomBase):
    pass


class DicomSeries:
    """A series of DicomImages within the same study set.
    """


class DicomStructure(DicomBase):
    pass


class DicomPlan(DicomBase):
    pass


class DicomStudy:
    """A DICOM study can hold one DICOM CT, one DICOM structure, one
    DICOM plan and one DICOM dose. These DICOM files must align via UID
    declaration.

    Not all types are required to create a DicomStudy.
    """


class DicomCollection:
    """A DicomCollection can hold an arbitrary number of DicomStudy objects.
    """



# Copyright (C) 2019 Cancer Care Associates

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.


import re
from copy import deepcopy

import pydicom

from pymedphys_utilities.libutils import get_imports

from .._level1.create import dicom_dataset_from_dict

IMPORTS = get_imports(globals())


def adjust_machine_name(dicom_dataset, new_machine_name):
    """Change the machine name within the DICOM header
    """

    new_dicom_dataset = deepcopy(dicom_dataset)

    for beam in new_dicom_dataset.BeamSequence:
        beam.TreatmentMachineName = new_machine_name

    return new_dicom_dataset


def adjust_machine_name_cli(args):
    dicom_dataset = pydicom.read_file(args.input_file, force=True)
    new_dicom_dataset = adjust_machine_name(
        dicom_dataset, args.new_machine_name)

    pydicom.write_file(args.output_file, new_dicom_dataset)


def delete_sequence_item_with_matching_key(sequence, key, value):
    new_sequence = deepcopy(sequence)

    for i, item in reversed(list(enumerate(sequence))):
        try:
            if value == getattr(item, key):
                new_sequence.pop(i)
        except AttributeError:
            pass

    return new_sequence


def adjust_rel_elec_density(dicom_dataset, adjustment_map,
                            ignore_missing_structure=False):
    """Append or adjust relative electron densities of structures
    """

    new_dicom_dataset = deepcopy(dicom_dataset)

    ROI_name_to_number_map = {
        structure_set.ROIName: structure_set.ROINumber
        for structure_set in new_dicom_dataset.StructureSetROISequence
    }

    ROI_number_to_observation_map = {
        observation.ReferencedROINumber: observation
        for observation in new_dicom_dataset.RTROIObservationsSequence
    }

    for structure_name, new_red in adjustment_map.items():
        try:
            ROI_number = ROI_name_to_number_map[structure_name]
        except KeyError:
            if ignore_missing_structure:
                continue
            else:
                raise

        observation = ROI_number_to_observation_map[ROI_number]

        try:
            physical_properties = observation.ROIPhysicalPropertiesSequence
        except AttributeError:
            physical_properties = []

        physical_properties = delete_sequence_item_with_matching_key(
            physical_properties, 'ROIPhysicalProperty', 'REL_ELEC_DENSITY')

        physical_properties.append(
            dicom_dataset_from_dict({
                'ROIPhysicalProperty': 'REL_ELEC_DENSITY',
                'ROIPhysicalPropertyValue': new_red
            })
        )

        observation.ROIPhysicalPropertiesSequence = physical_properties

    return new_dicom_dataset


def adjust_RED_cli(args):
    adjustment_map = {
        key: item
        for key, item in zip(
            args.adjustment_map[::2], args.adjustment_map[1::2])
    }

    dicom_dataset = pydicom.read_file(args.input_file, force=True)
    new_dicom_dataset = adjust_rel_elec_density(
        dicom_dataset, adjustment_map,
        ignore_missing_structure=args.ignore_missing_structure)

    pydicom.write_file(args.output_file, new_dicom_dataset)


def RED_adjustment_map_from_structure_names(structure_names):
    structure_name_containing_RED_regex = r'^.*RED\s*[=:]\s*(\d+\.?\d*)\s*$'
    pattern = re.compile(
        structure_name_containing_RED_regex, flags=re.IGNORECASE)

    adjustment_map = {
        structure: float(pattern.match(structure).group(1))
        for structure in structure_names
        if pattern.match(structure)
    }

    return adjustment_map


def adjust_RED_by_structure_name(dicom_dataset):
    """Adjust the structure electron density based on structure name.
    """
    structure_names = [
        structure_set.ROIName
        for structure_set in dicom_dataset.StructureSetROISequence
    ]

    adjustment_map = RED_adjustment_map_from_structure_names(structure_names)

    adjusted_dicom_dataset = adjust_rel_elec_density(
        dicom_dataset, adjustment_map)

    return adjusted_dicom_dataset


def adjust_RED_by_structure_name_cli(args):
    dicom_dataset = pydicom.read_file(args.input_file, force=True)
    new_dicom_dataset = adjust_RED_by_structure_name(dicom_dataset)

    pydicom.write_file(args.output_file, new_dicom_dataset)



# Copyright (C) 2016-2019 Matthew Jennings and Simon Biggs

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.


"""A DICOM RT Dose toolbox"""

import warnings

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import path

from scipy.interpolate import RegularGridInterpolator

import pydicom
import pydicom.uid

from .._level1.structure import pull_structure

from pymedphys_utilities.libutils import get_imports
IMPORTS = get_imports(globals())

# pylint: disable=C0103


def convert_xyz_to_dicom_coords(xyz_tuple):
    ZZ, YY, XX = np.meshgrid(
        xyz_tuple[2], xyz_tuple[1], xyz_tuple[0], indexing='ij')

    coords = np.array((XX, YY, ZZ), dtype=np.float64)
    return coords


def load_dose_from_dicom(ds, set_transfer_syntax_uid=True, reshape=True):
    r"""Extract the dose grid from a DICOM RT Dose file.

    .. deprecated:: 0.5.0
            `load_dose_from_dicom` will be removed in a future version of PyMedPhys.
            It is replaced by `extract_dose`, which provides additional dose-related
            information and conforms to a new coordinate system handling convention.
    """

    if set_transfer_syntax_uid:
        ds.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian

    if reshape:
        warnings.warn((
            '`load_dose_from_dicom` currently reshapes the dose grid. In a '
            'future version this will no longer occur. To begin using this '
            'function without the reshape pass the parameter `reshape=False` '
            'when calling `load_dose_from_dicom`.'), UserWarning)
        pixels = np.transpose(
            ds.pixel_array, (1, 2, 0))
    else:
        pixels = ds.pixel_array

    dose = pixels * ds.DoseGridScaling

    return dose


def load_xyz_from_dicom(ds):
    r"""Extract the coordinates of a DICOM RT Dose file's dose grid.

    .. deprecated:: 0.5.0
            `load_xyz_from_dicom` will be removed in a future version of PyMedPhys.
            It is replaced by `extract_dicom_patient_coords`, `extract_iec_patient_coords`
            and `extract_iec_fixed_coords`, which explicitly work in their respective
            coordinate systems.
    """

    warnings.warn((
        '`load_xyz_from_dicom` returns x, y & z values in the DICOM patient '
        'coordinate system and presumes the patient\'s orientation is HFS. '
        'This presumption may not be correct and so the function may return '
        'incorrect x, y, z values. In the future, this function will be removed. '
        'It is currently preserved for temporary backwards compatibility.'
    ), UserWarning)

    resolution = np.array(ds.PixelSpacing).astype(float)

    dx = resolution[0]
    x = (ds.ImagePositionPatient[0] + np.arange(0, ds.Columns * dx, dx))

    dy = resolution[1]
    y = (ds.ImagePositionPatient[1] + np.arange(0, ds.Rows * dy, dy))

    z = (np.array(ds.GridFrameOffsetVector) + ds.ImagePositionPatient[2])

    return x, y, z


def extract_iec_patient_xyz(ds):
    r"""Returns the x, y and z coordinates of a DICOM RT Dose file's dose grid
    in the IEC patient coordinate system

    Parameters
    ----------
    ds
        A `pydicom Dataset` - ordinarily returned by `pydicom.dcmread()`.
        Must represent a valid DICOM RT Dose file.

    Returns
    -------
    (x, y, z)
        A tuple containing three `ndarrays` corresponding to the `x`,
        `y` and `z` coordinates of the DICOM RT Dose file's dose grid in
        the IEC patient coordinate system [1]_:

        `x` corresponds to the patient's left-right axis and increases toward the
        left hand side of the patient.

        `y` corresponds to the patient's superoinferior axis and increases toward
        the head of the patient.

        `z` corresponds to the patient's anteroposterior axis and increases to the
        anterior side of the patient.

    Notes
    -----
    Supported scan orientations [2]_:

    =========================== =======================
    Orientation                 ImageOrientationPatient
    =========================== =======================
    Feet First Decubitus Left   [0, 1, 0, 1, 0, 0]
    Feet First Decubitus Right  [0, -1, 0, -1, 0, 0]
    Feet First Prone            [1, 0, 0, 0, -1, 0]
    Feet First Supine           [-1, 0, 0, 0, 1, 0]
    Head First Decubitus Left   [0, -1, 0, 1, 0, 0]
    Head First Decubitus Right  [0, 1, 0, -1, 0, 0]
    Head First Prone            [-1, 0, 0, 0, -1, 0]
    Head First Supine           [1, 0, 0, 0, 1, 0]
    =========================== =======================

    References
    ----------
    .. [1] "IEC 61217:2.0 Radiotherapy equipment – Coordinates, movements and scales"

    .. [2] O. McNoleg, "Generalized coordinate transformations for Monte Carlo
       (DOSXYZnrc and VMC++) verifications of DICOM compatible radiotherapy
       treatment plans", arXiv:1406.0014, Table 1,
       https://arxiv.org/ftp/arxiv/papers/1406/1406.0014.pdf
    """

    if ds.Modality != "RTDOSE":
        raise ValueError("The input DICOM file is not an RT Dose file")

    position = np.array(ds.ImagePositionPatient)
    orientation = np.array(ds.ImageOrientationPatient)

    di = float(ds.PixelSpacing[0])
    dj = float(ds.PixelSpacing[1])

    is_prone_or_supine = np.array_equal(
        np.absolute(orientation), np.array([1., 0., 0., 0., 1., 0.]))

    is_decubitus = np.array_equal(
        np.absolute(orientation), np.array([0., 1., 0., 1., 0., 0.]))

    if is_prone_or_supine:
        xflip = (orientation[0] == -1)
        zflip = (orientation[4] == 1)
        head_first = (xflip != zflip)

        x = orientation[0]*position[0] + np.arange(0, ds.Columns * di, di)
        z = -(orientation[4]*position[1] + np.arange(0, ds.Rows * dj, dj))

    elif is_decubitus:
        xflip = (orientation[3] == -1)
        zflip = (orientation[1] == 1)
        head_first = (xflip == zflip)

        x = -(orientation[3]*position[0] + np.arange(0, ds.Rows * dj, dj))
        z = orientation[1]*position[1] + np.arange(0, ds.Columns * di, di)
    else:
        raise ValueError(
            "Dose grid orientation is not supported. "
            "Dose grid slices must be aligned along the superoinferior axis of patient")

    if xflip:
        x = np.flip(x)
    if zflip:
        z = np.flip(z)

    if head_first:
        y = position[2] + np.array(ds.GridFrameOffsetVector)
    else:
        y = np.flip(-position[2] + np.array(ds.GridFrameOffsetVector))

    return (x, y, z)


def extract_iec_fixed_xyz(ds):
    r"""Returns the x, y and z coordinates of a DICOM RT Dose file's dose grid
    in the IEC fixed coordinate system.

    Parameters
    ----------
    ds
        A `pydicom Dataset` - ordinarily returned by `pydicom.dcmread()`.
        Must represent a valid DICOM RT Dose file.

    Returns
    -------
    (x, y, z)
        A tuple containing three `ndarrays` corresponding to the `x`,
        `y` and `z` coordinates of the DICOM RT Dose file's dose grid in
        the IEC fixed coordinate system [3]_:

            `x`
                An `np.ndarray` of coordinates on the horizontal plane that runs
                orthogonal to the gantry axis of rotation. The x axis increases
                towards a viewer's right hand side when facing the gantry standing
                at the isocentre. The isocentre has an x value of 0.
            `y`
                An `np.ndarray` of coordinates on the horizontal plane that runs
                parallel to the gantry axis of rotation. The positive direction
                of the y axis travels from the isocentre to the gantry. The
                isocentre has a y value of 0.
            `z`
                An `np.ndarray` of coordinates aligned vertically and passing through
                the isocentre. The positive direction of the z axis travels upwards
                and the isocentre has a z value of 0.

    Notes
    -----
    Supported scan orientations [4]_:

    =========================== =======================
    Orientation                 ImageOrientationPatient
    =========================== =======================
    Feet First Decubitus Left   [0, 1, 0, 1, 0, 0]
    Feet First Decubitus Right  [0, -1, 0, -1, 0, 0]
    Feet First Prone            [1, 0, 0, 0, -1, 0]
    Feet First Supine           [-1, 0, 0, 0, 1, 0]
    Head First Decubitus Left   [0, -1, 0, 1, 0, 0]
    Head First Decubitus Right  [0, 1, 0, -1, 0, 0]
    Head First Prone            [-1, 0, 0, 0, -1, 0]
    Head First Supine           [1, 0, 0, 0, 1, 0]
    =========================== =======================

    References
    ----------
    .. [3] "IEC 61217:2.0 Radiotherapy equipment – Coordinates, movements and scales"

    .. [4] O. McNoleg, "Generalized coordinate transformations for Monte Carlo
       (DOSXYZnrc and VMC++) verifications of DICOM compatible radiotherapy
       treatment plans", arXiv:1406.0014, Table 1,
       https://arxiv.org/ftp/arxiv/papers/1406/1406.0014.pdf
    """

    if ds.Modality != "RTDOSE":
        raise ValueError("The input DICOM file is not an RT Dose file")

    position = np.array(ds.ImagePositionPatient)
    orientation = np.array(ds.ImageOrientationPatient)

    di = float(ds.PixelSpacing[0])
    dj = float(ds.PixelSpacing[1])

    is_prone_or_supine = np.array_equal(
        np.absolute(orientation), np.array([1., 0., 0., 0., 1., 0.]))

    is_decubitus = np.array_equal(
        np.absolute(orientation), np.array([0., 1., 0., 1., 0., 0.]))

    # Only proceed if the DICOM RT Dose file has a supported orientation.
    # I.e. no pitch, yaw or non-cardinal roll angle exists between the dose
    # grid and the 'patient'
    if is_prone_or_supine:
        y_orientation = 2*(orientation[0] == orientation[4])-1

        x = orientation[0]*position[0] + np.arange(0, ds.Columns * di, di)
        z = -np.flip(orientation[4]*position[1] +
                     np.arange(0, ds.Rows * dj, dj))

    elif is_decubitus:
        y_orientation = 2*(orientation[1] != orientation[3])-1

        z = -np.flip(orientation[3]*position[0] +
                     np.arange(0, ds.Rows * dj, dj))
        x = orientation[1]*position[1] + np.arange(0, ds.Columns * di, di)
    else:
        raise ValueError(
            "Dose grid orientation is not supported. "
            "Dose grid slices must be aligned along the superoinferior axis of patient")

    y = y_orientation*position[2] + np.array(ds.GridFrameOffsetVector)

    return (x, y, z)


def extract_dicom_patient_xyz(ds):
    r"""Returns the x, y and z coordinates of a DICOM RT Dose file's dose grid
    in the DICOM patient coordinate system

    Parameters
    ----------
    ds
        A `pydicom Dataset` - ordinarily returned by `pydicom.dcmread()`.
        Must represent a valid DICOM RT Dose file.

    Returns
    -------
    (x, y, z)
        A tuple containing three `ndarrays` corresponding to the `x`,
        `y` and `z` coordinates of the DICOM RT Dose file's dose grid in
        the DICOM patient coordinate system [5]_:

        `x` corresponds to the patient's left-right axis and increases to the
        left hand side of the patient.

        `y` corresponds to the patient's anteroposterior axis and increases toward
        the posterior side of the patient.

        `z` corresponds to the patient's superoinferior axis and increases
        toward the head of the patient.

    Notes
    -----
    Supported scan orientations [6]_:

    =========================== =======================
    Orientation                 ImageOrientationPatient
    =========================== =======================
    Feet First Decubitus Left   [0, 1, 0, 1, 0, 0]
    Feet First Decubitus Right  [0, -1, 0, -1, 0, 0]
    Feet First Prone            [1, 0, 0, 0, -1, 0]
    Feet First Supine           [-1, 0, 0, 0, 1, 0]
    Head First Decubitus Left   [0, -1, 0, 1, 0, 0]
    Head First Decubitus Right  [0, 1, 0, -1, 0, 0]
    Head First Prone            [-1, 0, 0, 0, -1, 0]
    Head First Supine           [1, 0, 0, 0, 1, 0]
    =========================== =======================

    References
    ----------
    .. [5] "C.7.6.2.1.1 Image Position and Image Orientation",
       "DICOM PS3.3 2016a - Information Object Definitions",
       http://dicom.nema.org/MEDICAL/dicom/2016a/output/chtml/part03/sect_C.7.6.2.html#sect_C.7.6.2.1.1

    .. [6] O. McNoleg, "Generalized coordinate transformations for Monte Carlo
       (DOSXYZnrc and VMC++) verifications of DICOM compatible radiotherapy
       treatment plans", arXiv:1406.0014, Table 1,
       https://arxiv.org/ftp/arxiv/papers/1406/1406.0014.pdf
    """

    if ds.Modality != "RTDOSE":
        raise ValueError("The input DICOM file is not an RT Dose file")

    position = np.array(ds.ImagePositionPatient)
    orientation = np.array(ds.ImageOrientationPatient)

    di = float(ds.PixelSpacing[0])
    dj = float(ds.PixelSpacing[1])

    is_prone_or_supine = np.array_equal(
        np.absolute(orientation), np.array([1., 0., 0., 0., 1., 0.]))

    is_decubitus = np.array_equal(
        np.absolute(orientation), np.array([0., 1., 0., 1., 0., 0.]))

    # Only proceed if the DICOM RT Dose file has a supported orientation.
    # I.e. no pitch, yaw or non-cardinal roll angle exists between the dose
    # grid and the 'patient'
    if is_prone_or_supine:
        xflip = (orientation[0] == -1)
        yflip = (orientation[4] == -1)
        head_first = (xflip == yflip)

        x = orientation[0]*position[0] + np.arange(0, ds.Columns * di, di)
        y = orientation[4]*position[1] + np.arange(0, ds.Rows * dj, dj)

    elif is_decubitus:
        xflip = (orientation[3] == -1)
        yflip = (orientation[1] == -1)
        head_first = (xflip != yflip)

        x = orientation[3]*position[0] + np.arange(0, ds.Rows * dj, dj)
        y = orientation[1]*position[1] + np.arange(0, ds.Columns * di, di)
    else:
        raise ValueError(
            "Dose grid orientation is not supported. "
            "Dose grid slices must be aligned along the superoinferior axis of patient")

    if xflip:
        x = np.flip(x)
    if yflip:
        y = np.flip(y)

    if head_first:
        z = position[2] + np.array(ds.GridFrameOffsetVector)
    else:
        z = np.flip(-position[2] + np.array(ds.GridFrameOffsetVector))

    return (x, y, z)


def coords_and_dose_from_dicom(dicom_filepath):
    ds = pydicom.read_file(dicom_filepath, force=True)
    x, y, z = load_xyz_from_dicom(ds)
    coords = (y, x, z)
    dose = load_dose_from_dicom(ds)

    return coords, dose


def load_dicom_data(ds, depth_adjust):
    dose = load_dose_from_dicom(ds)
    crossplane, vertical, inplane = load_xyz_from_dicom(ds)

    depth = vertical + depth_adjust

    return inplane, crossplane, depth, dose


def arbitrary_profile_from_dicom_dose(ds, depth_adjust, inplane_ref, crossplane_ref, depth_ref):
    inplane, crossplane, depth, dose = load_dicom_data(ds, depth_adjust)

    interpolation_function = RegularGridInterpolator(
        (depth, crossplane, inplane), dose)
    points = [
        (a_depth_val, a_crossplane_val, an_inplane_val)
        for a_depth_val, a_crossplane_val, an_inplane_val
        in zip(depth_ref, crossplane_ref, inplane_ref)
    ]

    interpolated_dose = interpolation_function(points)

    return interpolated_dose


def extract_depth_dose(ds, depth_adjust, averaging_distance=0):
    inplane, crossplane, depth, dose = load_dicom_data(ds, depth_adjust)

    inplane_ref = abs(inplane) <= averaging_distance
    crossplane_ref = abs(crossplane) <= averaging_distance

    sheet_dose = dose[:, :, inplane_ref]
    column_dose = sheet_dose[:, crossplane_ref, :]

    depth_dose = np.mean(column_dose, axis=(1, 2))

    # uncertainty = np.std(column_dose, axis=(1, 2)) / depth_dose
    # assert np.all(uncertainty < 0.01),
    # "Shouldn't average over more than 1% uncertainty"

    return depth, depth_dose


def extract_profiles(ds, depth_adjust, depth_lookup, averaging_distance=0):

    inplane, crossplane, depth, dose = load_dicom_data(ds, depth_adjust)

    inplane_ref = abs(inplane) <= averaging_distance
    crossplane_ref = abs(crossplane) <= averaging_distance

    depth_reference = depth == depth_lookup

    dose_at_depth = dose[depth_reference, :, :]
    inplane_dose = np.mean(dose_at_depth[:, crossplane_ref, :], axis=(0, 1))
    crossplane_dose = np.mean(dose_at_depth[:, :, inplane_ref], axis=(0, 2))

    return inplane, inplane_dose, crossplane, crossplane_dose


def nearest_negative(diff):
    neg_diff = np.copy(diff)
    neg_diff[neg_diff > 0] = -np.inf
    return np.argmax(neg_diff)


def bounding_vals(test, values):
    npvalues = np.array(values).astype('float')
    diff = npvalues - test
    upper = nearest_negative(-diff)
    lower = nearest_negative(diff)

    return values[lower], values[upper]


def average_bounding_profiles(ds, depth_adjust, depth_lookup,
                              averaging_distance=0):
    inplane, crossplane, depth, _ = load_dicom_data(ds, depth_adjust)

    if depth_lookup in depth:
        return extract_profiles(
            ds, depth_adjust, depth_lookup, averaging_distance)
    else:
        print(
            'Specific depth not found, interpolating from surrounding depths')
        shallower, deeper = bounding_vals(depth_lookup, depth)

        _, shallower_inplane, _, shallower_crossplane = np.array(
            extract_profiles(ds, depth_adjust, shallower, averaging_distance))

        _, deeper_inplane, _, deeper_crossplane = np.array(
            extract_profiles(ds, depth_adjust, deeper, averaging_distance))

        depth_range = deeper - shallower
        shallower_weight = 1 - (depth_lookup - shallower) / depth_range
        deeper_weight = 1 - (deeper - depth_lookup) / depth_range

        inplane_dose = (
            shallower_weight * shallower_inplane +
            deeper_weight * deeper_inplane)
        crossplane_dose = (
            shallower_weight * shallower_crossplane +
            deeper_weight * deeper_crossplane)

        return inplane, inplane_dose, crossplane, crossplane_dose


def _get_index(z_list, z_val):
    indices = np.array([item[0] for item in z_list])
    # This will error if more than one contour exists on a given slice
    index = int(np.where(indices == z_val)[0])
    # Multiple contour sets per slice not yet implemented

    return index


def find_dose_within_structure(structure, dcm_struct, dcm_dose):
    x_dose, y_dose, z_dose = load_xyz_from_dicom(dcm_dose)
    dose = load_dose_from_dicom(dcm_dose)

    xx_dose, yy_dose = np.meshgrid(x_dose, y_dose)
    points = np.swapaxes(np.vstack([xx_dose.ravel(), yy_dose.ravel()]), 0, 1)

    x_structure, y_structure, z_structure = pull_structure(
        structure, dcm_struct)
    structure_z_values = np.array([item[0] for item in z_structure])

    structure_dose_values = np.array([])

    for z_val in structure_z_values:
        structure_index = _get_index(z_structure, z_val)
        dose_index = int(np.where(z_dose == z_val)[0])

        assert z_structure[structure_index][0] == z_dose[dose_index]

        structure_polygon = path.Path([
            (
                x_structure[structure_index][i],
                y_structure[structure_index][i]
            )
            for i in range(len(x_structure[structure_index]))
        ])
        mask = structure_polygon.contains_points(points).reshape(
            len(y_dose), len(x_dose))
        masked_dose = dose[:, :, dose_index]
        structure_dose_values = np.append(
            structure_dose_values, masked_dose[mask])

    return structure_dose_values


def create_dvh(structure, dcm_struct, dcm_dose):
    structure_dose_values = find_dose_within_structure(
        structure, dcm_struct, dcm_dose)
    hist = np.histogram(structure_dose_values, 100)
    freq = hist[0]
    bin_edge = hist[1]
    bin_mid = (bin_edge[1::] + bin_edge[:-1:])/2

    cumulative = np.cumsum(freq[::-1])
    cumulative = cumulative[::-1]
    bin_mid = np.append([0], bin_mid)

    cumulative = np.append(cumulative[0], cumulative)
    percent_cumulative = cumulative / cumulative[0] * 100

    plt.plot(bin_mid, percent_cumulative, label=structure)
    plt.title('DVH')
    plt.xlabel('Dose (Gy)')
    plt.ylabel('Relative Volume (%)')



# Copyright (C) 2018 Matthew Jennings, Simon Biggs

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version (the "AGPL-3.0+").

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License and the additional terms for more
# details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ADDITIONAL TERMS are also included as allowed by Section 7 of the GNU
# Affero General Public License. These additional terms are Sections 1, 5,
# 6, 7, 8, and 9 from the Apache License, Version 2.0 (the "Apache-2.0")
# where all references to the definition "License" are instead defined to
# mean the AGPL-3.0+.

# You should have received a copy of the Apache-2.0 along with this
# program. If not, see <http://www.apache.org/licenses/LICENSE-2.0>.

from glob import glob
from os.path import basename, dirname, join as pjoin
from copy import deepcopy

import numpy as np
import pydicom

from pymedphys_utilities.utilities import remove_file

from pymedphys_utilities.libutils import get_imports
from .._level1.constants import (
    BaselineDicomDictionary,
    BASELINE_KEYWORD_VR_DICT,
    DICOM_SOP_CLASS_NAMES_MODE_PREFIXES)

IMPORTS = get_imports(globals())


IDENTIFYING_KEYWORDS = ("AccessionNumber",
                        "AcquisitionDate",
                        "AcquisitionDateTime",
                        "AcquisitionTime",
                        "ContentCreatorName",
                        "ContentDate",
                        "ContentTime",
                        "CountryOfResidence",
                        "CurrentPatientLocation",
                        "CurveDate",
                        "CurveTime",
                        "Date",
                        "DateTime",
                        "EthnicGroup",
                        "InstanceCreationDate",
                        "InstanceCreationTime",
                        "InstanceCreatorUID",
                        "InstitutionAddress",
                        "InstitutionalDepartmentName",
                        "InstitutionName",
                        "IssuerOfPatientID",
                        "NameOfPhysiciansReadingStudy",
                        "OperatorsName",
                        "OtherPatientIDs",
                        "OtherPatientNames",
                        "OverlayDate",
                        "OverlayTime",
                        "PatientAddress",
                        "PatientAge",
                        "PatientBirthDate",
                        "PatientBirthName",
                        "PatientBirthTime",
                        "PatientID",
                        "PatientInstitutionResidence",
                        "PatientMotherBirthName",
                        "PatientName",
                        "PatientSex",
                        "PatientTelephoneNumbers",
                        "PerformingPhysicianIdentificationSequence",
                        "PerformingPhysicianName",
                        "PersonName",
                        "PhysiciansOfRecord",
                        "PhysiciansOfRecordIdentificationSequence",
                        "PhysiciansReadingStudyIdentificationSequence",
                        "ReferringPhysicianAddress",
                        "ReferringPhysicianIdentificationSequence",
                        "ReferringPhysicianName",
                        "ReferringPhysicianTelephoneNumbers",
                        "RegionOfResidence",
                        "ReviewerName",
                        "SecondaryReviewerName",
                        "SeriesDate",
                        "SeriesTime",
                        "StationName",
                        "StudyDate",
                        "StudyID",
                        "StudyTime",
                        "Time",
                        "VerifyingObserverName")


VR_ANONYMOUS_REPLACEMENT_VALUE_DICT = {'AS': "100Y",
                                       'CS': "ANON",
                                       'DA': "20190303",
                                       'DT': "20190303000900.000000",
                                       'LO': "Anonymous",
                                       'PN': "Anonymous",
                                       'SH': "Anonymous",
                                       'SQ': ["Anonymous"],
                                       'ST': "Anonymous",
                                       'TM': "000900.000000",
                                       'UI': "12345678"}


def anonymise_dataset(
        ds,
        replace_values=True,
        keywords_to_leave_unchanged=(),
        delete_private_tags=True,
        delete_unknown_tags=None,
        copy_dataset=True):
    r"""A simple tool to anonymise a DICOM dataset.

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The DICOM dataset to be anonymised.

    replace_values : bool, optional
        If set to `True`, DICOM tags will be anonymised using dummy
        "anonymous" values. This is often required for commercial
        software to successfully read anonymised DICOM files. If set to
        `False`, anonymised tags are simply given empty string values.
        Defaults to `True`.

    keywords_to_leave_unchanged : sequence, optional
        A sequence of DICOM keywords (corresponding to tags) to exclude
        from anonymisation. Private and unknown tags can be supplied.
        Empty by default.

    delete_private_tags : bool, optional
        A boolean to flag whether or not to remove all private
        (non-standard) DICOM tags from the DICOM file. These may also
        contain identifying information. Defaults to `True`.

    delete_unknown_tags : bool, pseudo-optional
        If left as the default value of `None` and `ds` contains tags
        that are not present in PyMedPhys` copy of `pydicom`'s DICOM
        dictionary, `anonymise_dataset()` will raise an error. The
        user must then either pass `True` or `False` to proceed. If set
        to `True`, all unrecognised tags that haven't been listed in
        `keywords_to_leave_unchanged` will be deleted. If set to
        `False`, these tags are simply ignored. Pass `False` with
        caution, since unrecognised tags may contain identifying
        information.

    copy_dataset : bool, optional
        If True, then a copy of `ds` is returned.

    Returns
    -------
    ds_anon : pydicom.dataset.Dataset
        An anonymised version of the input DICOM dataset.
    """

    if copy_dataset:
        ds_anon = deepcopy(ds)
    else:
        ds_anon = ds

    unknown_tags = unknown_tags_in_dicom_dataset(ds_anon)

    if delete_unknown_tags is None and unknown_tags:
        unknown_keywords = [ds_anon[tag].keyword for tag in unknown_tags]

        raise ValueError(
            "At least one of the non-private tags within your DICOM "
            "file is not within PyMedPhys's copy of the DICOM "
            "dictionary. It is possible that one or more of these tags "
            "contains identifying information. The unrecognised tags "
            "are:\n\n{}\n\nTo exclude these unknown tags from the "
            "anonymised DICOM dataset, pass `delete_unknown_tags=True` "
            "to this function. Any unknown tags passed to "
            "`tags_to_leave_unchanged` will not be deleted. If you'd "
            "like to ignore this error and keep all unknown tags in "
            "the anonymised DICOM dataset, pass "
            "`delete_unknown_tags=False` to this function. Finally, "
            "if you suspect that the PyMedPhys DICOM dictionary is out "
            "of date, please raise an issue on GitHub at "
            "https://github.com/pymedphys/pymedphys/issues."
            .format(unknown_keywords))

    elif delete_unknown_tags:
        unwanted_unknown_tags = []

        for tag in unknown_tags:
            if ds_anon[tag].keyword not in keywords_to_leave_unchanged:
                unwanted_unknown_tags.append(tag)
                del ds_anon[tag]

        for tag in unwanted_unknown_tags:
            if tag in ds_anon:
                raise AssertionError(
                    "Could not delete all unwanted, unknown tags.")

    if delete_private_tags:
        ds_anon.remove_private_tags()

    keywords_to_anonymise = _filter_identifying_keywords(
        keywords_to_leave_unchanged)

    ds_anon = _anonymise_tags(ds_anon, keywords_to_anonymise, replace_values)

    return ds_anon


def anonymise_file(
        dicom_filepath,
        delete_original_file=False,
        anonymise_filename=True,
        replace_values=True,
        keywords_to_leave_unchanged=(),
        delete_private_tags=True,
        delete_unknown_tags=None):
    r"""A simple tool to anonymise a DICOM file.

    Parameters
    ----------
    dicom_filepath : str
        The path to the DICOM file to be anonymised.

    delete_original_file : bool, optional
        If True and anonymisation completes successfully, then the
        original DICOM is deleted. Defaults to False.

    anonymise_filename : bool, optional
        If `True`, the DICOM filename is replaced by a filename of the
        form:

        "<2 char DICOM modality>.<SOP Instance UID>_Anonymised.dcm".

        E.g.: "RP.2.16.840.1.113669.[...]_Anonymised.dcm"

        This ensures that the filename contains no identifying
        information. If set to `False`, `anonymise_file()` simply
        appends "_Anonymised" to the original DICOM filename. Defaults
        to True.

    replace_values : bool, optional
        If set to `True`, DICOM tags will be anonymised using dummy
        "anonymous" values. This is often required for commercial
        software to successfully read anonymised DICOM files. If set to
        `False`, anonymised tags are simply given empty string values.
        Defaults to `True`.

    keywords_to_leave_unchanged : sequence, optional
        A sequence of DICOM keywords (corresponding to tags) to exclude
        from anonymisation. Private and unknown tags can be supplied.
        Empty by default.

    delete_private_tags : bool, optional
        A boolean to flag whether or not to remove all private
        (non-standard) DICOM tags from the DICOM file. These may
        also contain identifying information. Defaults to `True`.

    delete_unknown_tags : bool, pseudo-optional
        If left as the default value of `None` and `ds` contains tags
        that are not present in PyMedPhys` copy of `pydicom`'s DICOM
        dictionary, `anonymise_dataset()` will raise an error. The
        user must then either pass `True` or `False` to proceed. If set
        to `True`, all unrecognised tags that haven't been listed in
        `keywords_to_leave_unchanged` will be deleted. If set to
        `False`, these tags are simply ignored. Pass `False` with
        caution, since unrecognised tags may contain identifying
        information.
    """

    ds = pydicom.dcmread(dicom_filepath)

    if anonymise_filename:
        mode_prefix = DICOM_SOP_CLASS_NAMES_MODE_PREFIXES[ds.SOPClassUID.name]

        dicom_anon_filepath = pjoin(
            dirname(dicom_filepath),
            "{}.{}_Anonymised.dcm".format(mode_prefix, ds.SOPInstanceUID))
    else:
        basename_without_filetype = '.'.join(
            basename(dicom_filepath).split('.')[:-1])

        dicom_anon_filepath = pjoin(
            dirname(dicom_filepath),
            "{}_Anonymised.dcm".format(basename_without_filetype))

    ds_anon = anonymise_dataset(
        ds=ds,
        replace_values=replace_values,
        keywords_to_leave_unchanged=keywords_to_leave_unchanged,
        delete_private_tags=delete_private_tags,
        delete_unknown_tags=delete_unknown_tags,
        copy_dataset=False)

    ds_anon.save_as(dicom_anon_filepath)

    if delete_original_file:
        remove_file(dicom_filepath)

    return dicom_anon_filepath


def anonymise_directory(
        dicom_dirpath,
        delete_original_files=False,
        anonymise_filenames=True,
        replace_values=True,
        keywords_to_leave_unchanged=(),
        delete_private_tags=True,
        delete_unknown_tags=None):
    r"""A simple tool to anonymise all DICOM files in a directory and
    its subdirectories.

    Parameters
    ----------
    dicom_dirpath : str
        The path to the directory containing DICOM files to be
        anonymised.

    delete_original_files : bool, optional
        If True and anonymisation completes successfully, then the
        original DICOM files are deleted. Defaults to False.

    anonymise_filenames : bool, optional
        If True, the DICOM filenames are replaced by filenames of the
        form:

        "<2 char DICOM modality>.<SOP Instance UID>_Anonymised.dcm".

        E.g.: "RP.2.16.840.1.113669.[...]_Anonymised.dcm"

        This ensures that the filenames contain no identifying
        information. If `False`, `anonymise_directory()` simply
        appends "_Anonymised" to the original DICOM filenames. Defaults
        to True.

    replace_values : bool, optional
        If set to `True`, DICOM tags will be anonymised using dummy
        "anonymous" values. This is often required for commercial
        software to successfully read anonymised DICOM files. If set to
        `False`, anonymised tags are simply given empty string values.
        Defaults to `True`.

    keywords_to_leave_unchanged : sequence, optional
        A sequence of DICOM keywords (corresponding to tags) to exclude
        from anonymisation. Private and unknown tags can be supplied.
        Empty by default.

    delete_private_tags : bool, optional
        A boolean to flag whether or not to remove all private
        (non-standard) DICOM tags from the DICOM file. These may also
        contain identifying information. Defaults to `True`.

    delete_unknown_tags : bool, pseudo-optional
        If left as the default value of `None` and `ds` contains tags
        that are not present in PyMedPhys` copy of `pydicom`'s DICOM
        dictionary, `anonymise_dataset()` will raise an error. The
        user must then either pass `True` or `False` to proceed. If set
        to `True`, all unrecognised tags that haven't been listed in
        `keywords_to_leave_unchanged` will be deleted. If set to
        `False`, these tags are simply ignored. Pass `False` with
        caution, since unrecognised tags may contain identifying
        information.
    """
    dicom_filepaths = glob(dicom_dirpath + '/**/*.dcm', recursive=True)

    for dicom_filepath in dicom_filepaths:
        anonymise_file(
            dicom_filepath,
            delete_original_file=False,
            anonymise_filename=anonymise_filenames,
            replace_values=replace_values,
            keywords_to_leave_unchanged=keywords_to_leave_unchanged,
            delete_private_tags=delete_private_tags,
            delete_unknown_tags=delete_unknown_tags)

    # Separate loop provides the ability to raise Exceptions from the
    # unsuccessful deletion of the original DICOM files while preventing
    # these Exceptions from interrupting the batch anonymisation.
    if delete_original_files:
        for dicom_filepath in dicom_filepaths:
            remove_file(dicom_filepath)


# def anonymise_files_cli(args):


def is_anonymised_dataset(ds, ignore_private_tags=False):
    r"""Checks whether a DICOM dataset has been (fully) anonymised.
    """
    is_anonymised = True

    for elem in ds:
        if elem.keyword in IDENTIFYING_KEYWORDS:
            dummy_value = _get_anonymous_replacement_value(elem.keyword)
            if not (elem.value == '' or elem.value == dummy_value):
                is_anonymised = False
                # print(elem.value)
                break
        elif elem.tag.is_private and not ignore_private_tags:
            is_anonymised = False

    return is_anonymised


def is_anonymised_file(filepath, ignore_private_tags=False):
    r"""Checks whether a DICOM file has been (fully) anonymised.
    """
    ds = pydicom.dcmread(filepath)

    return is_anonymised_dataset(ds, ignore_private_tags=ignore_private_tags)


def is_anonymised_directory(dirpath, ignore_private_tags=False):
    r"""Checks whether all DICOM files in a directory have been (fully)
    anonymised.
    """
    is_anonymised = True
    dicom_filepaths = glob(dirpath + '/**/*.dcm', recursive=True)

    for dicom_filepath in dicom_filepaths:
        if not is_anonymised_file(dicom_filepath,
                                  ignore_private_tags=ignore_private_tags):
            is_anonymised = False
            # print(dicom_filepath)
            break

    return is_anonymised


def non_private_tags_in_dicom_dataset(ds):
    """Return all non-private tags from a DICOM dataset.
    """
    non_private_tags = [elem.tag for elem in ds if not elem.tag.is_private]
    return non_private_tags


def unknown_tags_in_dicom_dataset(ds):
    """Return all non-private tags from a DICOM dataset that do not
    exist in the PyMedPhys copy of the DICOM dictionary.
    """

    non_private_tags_in_dataset = np.array(
        non_private_tags_in_dicom_dataset(ds))

    are_non_private_tags_in_dict_baseline = [
        tag in BaselineDicomDictionary.keys()
        for tag in non_private_tags_in_dataset]

    unknown_tags = list(non_private_tags_in_dataset[
        np.invert(are_non_private_tags_in_dict_baseline)])

    return unknown_tags


def _anonymise_tags(ds_anon, keywords_to_anonymise, replace_values):
    """Anonymise all desired DICOM elements. 
    """
    for keyword in keywords_to_anonymise:
        if hasattr(ds_anon, keyword):
            if replace_values:
                replacement_value = _get_anonymous_replacement_value(keyword)
            else:
                replacement_value = ''
            setattr(ds_anon, keyword, replacement_value)

    return ds_anon


def _filter_identifying_keywords(keywords_to_leave_unchanged):
    r"""Removes DICOM keywords that the user desires to leave unchanged
    from the list of known DICOM identifying keywords and returns the
    resulting keyword list.
    """
    keywords_filtered = list(IDENTIFYING_KEYWORDS)
    for keyword in keywords_to_leave_unchanged:
        try:
            keywords_filtered.remove(keyword)
        except ValueError:
            # Value not in list. TODO: Warn?
            pass

    return keywords_filtered


def _get_anonymous_replacement_value(keyword):
    """Get an appropriate dummy anonymisation value for a DICOM element
    based on its value representation (VR)
    """
    vr = BASELINE_KEYWORD_VR_DICT[keyword]
    return VR_ANONYMOUS_REPLACEMENT_VALUE_DICT[vr]
